/* 
   uFS Project [ufs_fileio.c]
   (c) Simon Kueppers
   20.12.06 Germany

   This file is only made for file handling. Here you will find functions to 
   create, delete, write to or read from files.
  
   For function descriptions look at ufs_dirio.h
   

   PS:
   I tried to comment the code as best as i was able to. This is a little bit confusing
   for a german, so please excuse if something is not that right it should be.
   Have fun with this code but remember the BEER LICENSE when you copy+paste something
   from here :-)

*/

#include <string.h>									//strcpy()
#include "stdint.h"									//Vartypedefs
#include "ufs.h"									//function prototypes to ufs-core
#include "sd.h"										//function prototypes to access sd card
#include "time.h"									//function prototypes to real time clock
#include "ufs_fileio.h"								//function prototypes and defines
#include "ufs_dirio.h"								//function prototypes to dir management


/*
   This function (re)calculates the filesize of a file
   and puts this information into the headpage
*/
uint8_t ufs_fcalcsize(uint32_t fileadress)
{
	struct t_page TempPage;
	uint32_t tempadress;
	uint32_t nBytes=0;

	if (fileadress == 0)							//check if fileadress is valid
		return 0;

	sd_read(fileadress, &TempPage);					//read page to TempPage

	while((tempadress = TempPage.hdr_pNextpage) != 0)//while not last page in chain
	{
		sd_read(tempadress, &TempPage);				//read new page to TempPage
		nBytes += TempPage.hdr_nValidbytes;			//and calculate bytes in this page to bytecounter
	}

	sd_read(fileadress, &TempPage);					//read headpage again
	TempPage.headpage.filesize = nBytes;			//write size to headpage
	sd_write(fileadress, &TempPage);

	return 1;										//return successful
}

/*
   This function re-positions the filepointer to an given offset 'offset'
   relative to 'origin'
*/
uint8_t	ufs_fseek (struct t_file* pHandle, uint32_t offset, uint8_t origin)
{
	struct t_page TempPage;
	uint32_t tempadress=0;

	if (origin == FSEEK_START)						//apply 'offset' relative to start of file
	{
		sd_read(pHandle->pFileadress, &TempPage);	//read first page of file
		tempadress = TempPage.hdr_pNextpage;		//read next page
		sd_read(tempadress, &TempPage);

		pHandle->pCurrentPage = tempadress;		

		while (offset > 504)						//Are we now at searched page?
		{
			if (TempPage.hdr_pNextpage)				//if no, check if there is a next page to jump to
			{
				tempadress = TempPage.hdr_pNextpage;//If there is a next page, store adress of it
				sd_read(tempadress, &TempPage);		//Read this page to TempPage buffer
			} else									//If there is no next page anymore, return error
				return 0;

			offset -= 504;							//subtract 504 bytes
		}

		if (TempPage.hdr_nValidbytes >= offset)		//Is the searched byte in current page valid?
		{
			pHandle->pCurrentPage = tempadress;		//Store adress of current page
			pHandle->nCurrentByte = (uint16_t)offset;	//Store searched (and found) position in filehandle
			pHandle->pCurrentByte = (uint8_t*)(&pHandle->buffer.filepage.data) + (uint16_t)offset;

			sd_read(tempadress, &pHandle->buffer);
		}
		else										//Else
			return 0;								//Return error
		
	}else 
	if (origin == FSEEK_CUR)						//apply 'offset' relative to current position in file
	{
		sd_read(pHandle->pCurrentPage, &TempPage);

		while (offset > 504)						//Are we now at searched page?
		{
			if (TempPage.hdr_pNextpage)				//If yes, check if theres a next page
			{
				tempadress = TempPage.hdr_pNextpage;//If there is a next page, store adress of it
				sd_read(tempadress, &TempPage);		//Read this sector to TempPage buffer
			} else									//If there is no next page anymore, return 0
				return 0;

			offset -= 504;							//subtract 504 bytes
		}
		
		if (TempPage.hdr_nValidbytes >= offset)		//Is the searched position in current page valid?
		{			
			pHandle->pCurrentPage = tempadress;		//Store adress of current page
			pHandle->nCurrentByte = (uint16_t)offset;	//Store searched (and found) position in filehandle
			pHandle->pCurrentByte = (uint8_t*)(&pHandle->buffer.filepage.data) + (uint16_t)offset;

			sd_read(tempadress, &pHandle->buffer);
		}

		else										//Else
			return 0;								//Return error

		
	}else
	if (origin == FSEEK_END)						//this part is not implemented fully. it jumps always to the last byte in file + 1
	{
		sd_read(pHandle->pFileadress, &TempPage);
	
		if (TempPage.hdr_pNextpage == 0)
			return 0;

		while (TempPage.hdr_pNextpage != 0)
		{
			tempadress = TempPage.hdr_pNextpage;
			sd_read(tempadress, &TempPage);
		}
		
		sd_read(tempadress, &pHandle->buffer);

		pHandle->pCurrentPage = tempadress;
		pHandle->nCurrentByte = TempPage.hdr_nValidbytes;
		pHandle->pCurrentByte = (uint8_t*)(&pHandle->buffer.filepage.data) + TempPage.hdr_nValidbytes;

		
	}

	return 1;
}

/*
   This function deletes a file specified by 'path'
*/
uint8_t	ufs_fdelete	(uint8_t* path)
{
	struct t_page TempPage;
	uint32_t tempadress;
	uint32_t fileadress;


	fileadress = ufs_getfileadress(0, path);
	
	if (fileadress == 0) return 0;


	sd_read(fileadress, &TempPage);

	ufs_delentry(fileadress);
	ufs_pfree(fileadress);

	while((tempadress = TempPage.hdr_pNextpage) != 0)
	{
		sd_read(tempadress, &TempPage);

		ufs_pfree(tempadress);
	}

	ufs_writebitmapbuffer();

	return 1;


}

/*
   This function creates an empty file specified by 'path'
*/
uint32_t ufs_fcreate (uint8_t* path)
{
	uint32_t diradress, linkadress;
	struct t_page TempPage;
	uint8_t* pObjname = 0;
	uint8_t* pPath = path;
	uint8_t c;

	while ((c = *path) != 0)										//Scan Filename
	{
		if (c == '/')												//for each '/' its containing (or not containing)
			pObjname = path;										//set pointer to last found '/' in string
		
		path++;
	}

	if (pObjname == 0)												//There was no '/' in string
	{
		diradress = ROOTDIR_ADR;									//new file will be created in root directory
		pObjname = pPath;											//desired filename begins at array-element 0

	}else
	if (pObjname == pPath)											//There is a '/' at the beginning
	{
		diradress = ROOTDIR_ADR;									//create new file in root directory
		pObjname++;													//next char is the beginning of the filename

	}else
	if (*(pObjname + 1) == '\0')									//the last found '/' is the last char in string
	{
		return 0;													//sorry, no filename provided. exit with error
	
	}else															//string contains path & filename
	{
		*pObjname = '\0';
	
		diradress = ufs_getdiradress(ROOTDIR_ADR, pPath);			//get the directory adress of the scanned path

		if (diradress == 0)											//no (existent) directory
			return 0;												//return with error
		
		*pObjname = '/';											//replace string terminating zero with '/'
		pObjname++;													//pObjname now points to the beginning of the filename

	}

	
	if((linkadress = ufs_palloc()) == 0)							//allocate a new page
		return 0;													//and exist if not possible

	if (ufs_exists(diradress, pObjname))
		return 0;

	ufs_addentry(diradress, linkadress);							//add dir entry to scanned directory

	TempPage.headpage.attrflags = 0;								//configure headpage
	TempPage.headpage.filesize = 0;
	TempPage.headpage.tCreate =
	TempPage.headpage.tLastaccess =
	TempPage.headpage.tLastchange = gettime();
	TempPage.headpage.pParent = diradress;
	strcpy(TempPage.headpage.objname, pObjname);
	
	TempPage.hdr_nValidbytes = 504;
	TempPage.hdr_pNextpage = 0;
		
	sd_write(linkadress, &TempPage);								//and write it to allocated space

	ufs_writebitmapbuffer();										//synchronize bitmapbuffer

	return linkadress;

}

/*
   opens a file specified by 'path' to a filehandle 'pHandle' with mode 'mode'
*/
uint8_t ufs_fopen(struct t_file* pHandle, uint8_t* path, uint8_t mode)
{
	uint32_t fileadress;

	fileadress = ufs_getadress(ROOTDIR_ADR, path);					//get adress of file
	
	if (ufs_isdir(fileadress))										//check if file exist and is not a dir
		return 0;
		
	if (mode == 'w')												//Mode "write"
	{
		ufs_fdelete(path);											//delete file if exists
		fileadress = ufs_fcreate(path);								//creates empty file 

	}
	
	pHandle->pCurrentPage =
	pHandle->pFileadress = fileadress;								//configure parameters
	pHandle->nCurrentByte = 504;									//for fileio

	sd_read(fileadress, &pHandle->buffer);	

	if (mode == 'a')												//if mode was "append"
		ufs_fseek(pHandle, 0, FSEEK_END);							//set cursor to first new byte in file

						//and read first page to filebuffer

	return 1;

}

/*
   Closes an opened file with handle 'pHandle'
*/
void ufs_fclose (struct t_file* pHandle)
{
	if (pHandle->pCurrentPage)										//If CurrentPage Pointer is valid
	{
		sd_write(pHandle->pCurrentPage, &pHandle->buffer);			//write filebuffer back to disk
		ufs_fcalcsize(pHandle->pFileadress);						//calculate new filesize
		ufs_writebitmapbuffer();									//and sync bitmapbuffer
	}
}

/*
   Puts a char to a file with filehandle 'pHandle' and increments filepointer by one
*/
void ufs_fputc(struct t_file* pHandle, uint8_t b)
{
	uint32_t ptemp;

	if (pHandle->nCurrentByte >= 504)								//If there is no space left
	{
		if (pHandle->buffer.hdr_pNextpage)							//If there is a next page
		{
			sd_write(pHandle->pCurrentPage, &pHandle->buffer);		//write current buffered page to disk

			pHandle->pCurrentPage = pHandle->buffer.hdr_pNextpage;	//set adress of next page in filehandle

			sd_read(pHandle->buffer.hdr_pNextpage, &pHandle->buffer);//and read it to buffer

			pHandle->nCurrentByte = 0;								//start at byte 0
			pHandle->pCurrentByte = &pHandle->buffer.filepage.data;	//set pointer to byte 0 in buffer


		}else														//if there is no next page
		{
			if ((ptemp = ufs_palloc()) == 0)						//create one
				return;												//no space left? exit

			pHandle->buffer.hdr_pNextpage = ptemp;					//set pNextpage of current buffered page to new allocated page
			sd_write(pHandle->pCurrentPage, &pHandle->buffer);		//and write current buffered page to disk

			pHandle->pCurrentPage = ptemp;							//set adress of next page in filehandle
			pHandle->nCurrentByte = 0;								//start at byte 0
			pHandle->pCurrentByte = &pHandle->buffer.filepage.data;	//set pointer to byte 0 in buffer

			pHandle->buffer.hdr_pNextpage = 0;						//there is no next page after new page
			pHandle->buffer.hdr_nValidbytes = 0;					//there are no valid bytes in here

		}
	}

	*((uint8_t*)pHandle->pCurrentByte) = b;							//write current byte through pointer in filehandle
	if (pHandle->nCurrentByte >= pHandle->buffer.hdr_nValidbytes)	//if written char is not implied in pagesize
		pHandle->buffer.hdr_nValidbytes++;							//increment pagesize
	pHandle->nCurrentByte++;
	((uint8_t*)pHandle->pCurrentByte)++;

}

/*
   Puts a '\0' terminated string to a file and increments filepointer by strlen(s)
*/
void ufs_fputs(struct t_file* pHandle, uint8_t* s)
{
	uint8_t c;

	while((c = *s++) != '\0')										//if current char is not 0
		ufs_fputc(pHandle, c);										//write char

}

/*
   Gets a char from file and increments filepointer by one
*/
uint8_t ufs_fgetc(struct t_file* pHandle)
{
	uint8_t temp;

	if(pHandle->pCurrentPage == 0)									//If file has no content, return 0 (eof)
		return 0;
	
	if (pHandle->nCurrentByte >= pHandle->buffer.hdr_nValidbytes)	//Is going-to-read byte in current page?
	{
		if (pHandle->buffer.hdr_pNextpage == 0)
			return 0;												//if eof, return error

		sd_write(pHandle->pCurrentPage, &pHandle->buffer);			//write current buffered page to disk

		pHandle->pCurrentPage = pHandle->buffer.hdr_pNextpage;		//set adress of next page in filehandle

		sd_read(pHandle->buffer.hdr_pNextpage, &pHandle->buffer);	//and read it to buffer

		pHandle->nCurrentByte = 0;									//start at byte 0
		pHandle->pCurrentByte = &pHandle->buffer.filepage.data;		//set pointer to byte 0 in buffer

	}

	temp = *((uint8_t*)pHandle->pCurrentByte);						//read byte through pointer

	pHandle->nCurrentByte++;										//go to next byte
	((uint8_t*)pHandle->pCurrentByte)++;

	return temp;													//return read byte;
}

/*
   Checks if next fgetX will lead to an eof. it checks if there are still bytes left to read.
*/
uint8_t ufs_feof(struct t_file* pHandle)
{
	if (pHandle->pCurrentPage == 0)									//If file has no content, return 0 (eof)
		return 1;
	
	if (pHandle->nCurrentByte >= pHandle->buffer.hdr_nValidbytes)	//if next read-access will try to load a new page to buffer
		if (pHandle->buffer.hdr_pNextpage == 0)						//and if there is no next page
			return 1;												//return TRUE -> file is at eof

	return 0;														//everything is ok
}

/*
   gets size of file specified by 'path'
*/
uint32_t ufs_fgetsize (uint8_t* path)
{
	struct t_page TempPage;
	uint32_t fileadress;

	fileadress = ufs_getfileadress(ROOTDIR_ADR, path);				//get file adress

	if (fileadress == 0)											//if it does not exist or is a dir, return error
		return 0;

	sd_read(fileadress, &TempPage);									//read page at fileadress (will be a headpage)

	return TempPage.headpage.filesize;								//and return filesize-attribute
}

/* Not implemented yet
uint8_t	ufs_fcopy(uint8_t* srcpath, uint8_t* destpath);
uint8_t ufs_fread(struct t_file* ppHandle, void* pBuf, uint16_t nBytes);
uint8_t ufs_fwrite(struct t_file* ppHandle, void* pBuf, uint16_t nBytes);
*/